# coding=utf-8
import os

from tf_geometric import SparseAdj
from tf_geometric.nn import gcn_norm_adj, gcn_build_cache_for_graph
from tf_sparse import SparseMatrix

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
from tf_geometric.utils import tf_utils
import tensorflow as tf
import tf_geometric as tfg
from tqdm import tqdm
import time
import tf_sparse as tfs
import warnings
# tf.config.run_functions_eagerly(True)





def gcn(x, edge_index, edge_weight, kernel, bias=None, activation=None,
        renorm=True, improved=False, cache=None):
    """
    Functional API for Graph Convolutional Networks.

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param kernel: Tensor, shape: [num_features, num_output_features], weight
    :param bias: Tensor, shape: [num_output_features], bias
    :param activation: Activation function to use.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        To use @tf_utils.function with gcn, you should cache the noremd edge information before the first call of the gcn.

        - (1) If you're using OOP APIs tfg.layers.GCN:

              gcn_layer.build_cache_for_graph(graph)

        - (2) If you're using functional API tfg.nn.gcn:

              from tf_geometric.nn.conv.gcn import gcn_build_cache_for_graph
              gcn_build_cache_for_graph(graph)

    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tfs.shape(x)[0]

    sparse_adj = SparseAdj(edge_index, edge_weight, [num_nodes, num_nodes])
    normed_sparse_adj = gcn_norm_adj(sparse_adj, renorm, improved, cache)

    # SparseTensor is usually used for one-hot node features (For example, feature-less nodes.)
    if isinstance(x, tf.sparse.SparseTensor):
        h = tf.sparse.sparse_dense_matmul(x, kernel)
    else:
        h = x @ kernel

    h = normed_sparse_adj @ h

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h



class GCN(tf.keras.Model):
    """
    Graph Convolutional Layer
    """

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def __init__(self, units, activation=None,
                 use_bias=True,
                 renorm=True, improved=False,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        self.activation = activation
        self.use_bias = use_bias

        self.kernel = None
        self.bias = None

        self.renorm = renorm
        self.improved = improved

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph, self.renorm, self.improved, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn("'GCN.cache_normed_edge(graph, override)' is deprecated, use 'GCN.build_cache_for_graph(graph, override)' instead", DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gcn(x, edge_index, edge_weight, self.kernel, self.bias,
                   activation=self.activation, renorm=self.renorm, improved=self.improved, cache=cache)















dataset = "ogbn-arxiv"
graph, (train_index, valid_index, test_index) = tfg.datasets.OGBNodePropPredDataset(dataset).load_data()
# graph: tfg.Graph

num_classes = graph.y.max() + 1
# graph.convert_data_to_tensor()

drop_rate = 0.5
learning_rate = 1e-2


# Multi-layer GCN Model
class GCNModel(tf.keras.Model):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gcn0 = GCN(256, activation=tf.nn.relu)
        self.gcn1 = GCN(256, activation=tf.nn.relu)
        self.gcn2 = GCN(num_classes)
        self.dropout = tf.keras.layers.Dropout(drop_rate)

    def call(self, inputs, training=None, mask=None, cache=None):
        # start = tf.timestamp()

        x, edge_index, edge_weight = inputs
        # h = self.dropout(x, training=training)
        h = self.gcn0([x, edge_index, edge_weight], cache=cache)
        h = self.dropout(h, training=training)

        h = self.gcn1([h, edge_index, edge_weight], cache=cache)
        h = self.dropout(h, training=training)
        h = self.gcn2([h, edge_index, edge_weight], cache=cache)

        # tf.print("forward time: ", tf.timestamp() - start)

        return h

# Multi-layer GCN Model
# class GCNModel1(tf.keras.Model):
#
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.gcn0 = tfg.layers.GCN(256, activation=tf.nn.relu)
#         self.gcn1 = tf.keras.layers.Dense(256, activation=tf.nn.relu)
#         self.gcn2 = tf.keras.layers.Dense(num_classes)
#         self.dropout = tf.keras.layers.Dropout(drop_rate)
#
#     def call(self, inputs, training=None, mask=None, cache=None):
#         start = tf.timestamp()
#
#         x, edge_index, edge_weight = inputs
#
#         h = self.gcn1(x, training=training)
#
#         adj: SparseAdj = cache["gcn_normed_adj_True_False"]
#         h = adj @ h
#
#         h = self.gcn2(h, training=training)
#
#
#
#         tf.print("forward time: ", tf.timestamp() - start)
#
#         return h

model = GCNModel()


# @tf_utils.function can speed up functions for TensorFlow 2.x.
# @tf_utils.function is not compatible with TensorFlow 1.x and dynamic graph.cache.
# @tf_utils.function
def forward(graph, training=False):
    return model([graph.x, graph.edge_index, graph.edge_weight], training=training, cache=graph.cache)




# The following line is only necessary for using GCN with @tf_utils.function
# For usage without @tf_utils.function, you can commont the following line and GCN layers can automatically manager the cache
model.gcn0.build_cache_for_graph(graph)



# edge_index, edge_weight, shape = graph.cache[list(graph.cache.keys())[0]]
# print(edge_index.shape)
# adsfafsd

# print(tf.constant(adj))
# asdfasdf

# adj.index = adj.edge_index.numpy()
# adj.value = adj.edge_weight.numpy()

# adj.index = tf.constant(adj.edge_index)
# adj.value = tf.constant(adj.edge_weight)



# @tf_utils.function
def compute_loss(logits, mask_index, vars):
    # start = tf.timestamp()

    masked_logits = tf.gather(logits, mask_index)
    masked_labels = tf.gather(graph.y, mask_index)
    # losses = tf.nn.softmax_cross_entropy_with_logits(
    #     logits=masked_logits,
    #     labels=tf.one_hot(masked_labels, depth=num_classes)
    # )

    losses = tf.nn.sparse_softmax_cross_entropy_with_logits(
        logits=masked_logits,
        labels=masked_labels
    )



    kernel_vars = [var for var in vars if "kernel" in var.name]
    l2_losses = [tf.nn.l2_loss(kernel_var) for kernel_var in kernel_vars]

    loss = tf.reduce_mean(losses) + tf.add_n(l2_losses) * 1e-4

    # tf.print("loss time: ", tf.timestamp() - start)

    return loss


optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)


@tf_utils.function
# @tf_utils.function(experimental_compile=True)
def train_step():

    print("retracing ......")

    # logits = forward(graph, training=True)
    # loss = logits[0][0]#compute_loss(logits, train_index, model.trainable_variables)




    with tf.GradientTape() as tape:
        logits = forward(graph, training=True)
        loss = compute_loss(logits, train_index, tape.watched_variables())

    vars = tape.watched_variables()
    grads = tape.gradient(loss, vars)
    optimizer.apply_gradients(zip(grads, vars))


    return loss


# func = train_step.get_concrete_function()
# graph = func.graph
# for node in graph.as_graph_def().node:
#   print(f'{node.input} -> {node.name}')
#
#
# print("========")

@tf_utils.function
# @tf_utils.function(experimental_compile=True)
def evaluate():
    logits = forward(graph)
    masked_logits = tf.gather(logits, test_index)
    masked_labels = tf.gather(graph.y, test_index)

    y_pred = tf.argmax(masked_logits, axis=-1, output_type=tf.int32)

    corrects = tf.equal(y_pred, masked_labels)
    accuracy = tf.reduce_mean(tf.cast(corrects, tf.float32))
    return accuracy


loss = train_step()

# train_step_ = train_step.get_concrete_function()
# graph = train_step_.graph
# # for node in graph.as_graph_def().node:
# #   print(f'{node.input} -> {node.name}')
# #
# # print("======")
# # print(graph.as_graph_def())
#
# with open("numpy.txt", "w", encoding="utf-8") as f:
#     f.write(str(graph.as_graph_def()))
# exit(0)

accuracy = evaluate()

start_time = time.time()

print(train_step.pretty_printed_concrete_signatures())

for step in range(1, 20001):
    loss = train_step()
    if step % 20 == 0:
        accuracy = evaluate()
        # accuracy = 0.0
        print("step = {}\tloss = {}\taccuracy = {}\tavg_time = {}".format(step, loss, accuracy, (time.time() - start_time) / step))




print("\nstart speed test...")
num_test_iterations = 1000
start_time = time.time()
for _ in tqdm(range(num_test_iterations)):
    logits = forward(graph)
end_time = time.time()
print("mean forward time: {} seconds".format((end_time - start_time) / num_test_iterations))

if tf.__version__[0] == "1":
    print("** @tf_utils.function is disabled in TensorFlow 1.x. "
          "Upgrade to TensorFlow 2.x for 10X faster speed. **")


# coding=utf-8

import warnings
import tensorflow as tf
from tf_geometric.nn.conv.chebynet import chebynet, chebynet_cache_normed_edge


class ChebyNet(tf.keras.Model):
    """
    The chebyshev spectral graph convolutional operator from the
    `"Convolutional Neural Networks on Graphs with Fast Localized Spectral
    Filtering" <https://arxiv.org/abs/1606.09375>`_ paper

    """

    def __init__(self, units, k, activation=None, use_bias=True, normalization_type="sym",
                 use_dynamic_lambda_max=False,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param k: Chebyshev filter size (default: '3").
        :param lambda_max:
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param activation: Activation function to use.
        :param normalization_type: The normalization scheme for the graph
            Laplacian (default: :obj:`"sym"`)
        :param use_dynamic_lambda_max: If true, compute max eigen value for each forward,
            otherwise use 2.0 as the max eigen value
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        assert k > 0
        assert normalization_type in [None, 'sym', 'rw'], 'Invalid normalization'

        self.k = k

        self.use_bias = use_bias

        self.kernels = []
        self.bias = None

        self.activation = activation
        self.normalization_type = normalization_type
        self.use_dynamic_lambda_max = use_dynamic_lambda_max

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        for k in range(self.k):
            kernel = self.add_weight("kernel{}".format(k), shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
            self.kernels.append(kernel)

        # self.kernel = self.add_weight("kernel", shape=[self.K, num_features, self.units],
        #                               initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        chebynet_cache_normed_edge(graph, self.normalization_type,
                                   use_dynamic_lambda_max=self.use_dynamic_lambda_max, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn(
            "'ChebyNet.cache_normed_edge(graph, override)' is deprecated, use 'ChebyNet.build_cache_for_graph(graph, override)' instead",
            DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return chebynet(x, edge_index, edge_weight, self.k, self.kernels, self.bias, self.activation,
                        self.normalization_type, use_dynamic_lambda_max=self.use_dynamic_lambda_max, cache=cache)



# coding=utf-8

from tf_geometric.nn.conv.gcn import gcn, gcn_build_cache_for_graph
import tensorflow as tf
import warnings


class GCN(tf.keras.Model):
    """
    Graph Convolutional Layer
    """

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features, self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def __init__(self, units, activation=None,
                 use_bias=True,
                 renorm=True, improved=False,
                 edge_drop_rate=0.0,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param edge_drop_rate: Dropout rate of the propagation weights.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units

        self.activation = activation
        self.use_bias = use_bias

        self.edge_drop_rate = edge_drop_rate

        self.kernel = None
        self.bias = None

        self.renorm = renorm
        self.improved = improved

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph, self.renorm, self.improved, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn("'GCN.cache_normed_edge(graph, override)' is deprecated, use 'GCN.build_cache_for_graph(graph, override)' instead", DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return gcn(x, edge_index, edge_weight, self.kernel, self.bias,
                   activation=self.activation, renorm=self.renorm, improved=self.improved,
                   edge_drop_rate=self.edge_drop_rate,
                   training=training, cache=cache)



# coding=utf-8

import warnings
import tensorflow as tf
from tf_geometric.nn.conv.gcn import gcn_build_cache_for_graph

from tf_geometric.nn.conv.tagcn import tagcn


class TAGCN(tf.keras.Model):
    """
    The topology adaptive graph convolutional networks operator from the
     `"Topology Adaptive Graph Convolutional Networks"
     <https://arxiv.org/abs/1710.10370>`_ paper
    """

    def __init__(self, units, k=3, activation=None, use_bias=True,
                 renorm=False, improved=False,
                 kernel_regularizer=None, bias_regularizer=None,
                 *args, **kwargs):
        """

        :param units: Positive integer, dimensionality of the output space.
        :param k: Number of hops (default: '3").
        :param activation: Activation function to use.
        :param use_bias: Boolean, whether the layer uses a bias vector.
        :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
        :param improved: Whether use improved GCN or not.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrix.
        :param bias_regularizer: Regularizer function applied to the bias vector.
        """
        super().__init__(*args, **kwargs)
        self.units = units
        assert k > 0

        self.k = k

        self.activation = activation
        self.use_bias = use_bias

        self.kernel = None
        self.bias = None

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.renorm = renorm
        self.improved = improved

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        num_features = x_shape[-1]

        self.kernel = self.add_weight("kernel", shape=[num_features * (self.k + 1), self.units],
                                      initializer="glorot_uniform", regularizer=self.kernel_regularizer)
        if self.use_bias:
            self.bias = self.add_weight("bias", shape=[self.units],
                                        initializer="zeros", regularizer=self.bias_regularizer)

    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph, self.renorm, self.improved, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn("'TAGCN.cache_normed_edge(graph, override)' is deprecated, use 'TAGCN.build_cache_for_graph(graph, override)' instead", DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)


    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return tagcn(x, edge_index, edge_weight, self.k, self.kernel,
                     bias=self.bias, activation=self.activation, renorm=self.renorm,
                     improved=self.improved, cache=cache)



# coding=utf-8
import os

# Enable GPU 0
os.environ["CUDA_VISIBLE_DEVICES"] = "0"

import tf_geometric as tfg
import tensorflow as tf
import numpy as np
from tf_geometric.utils.graph_utils import convert_edge_to_directed

# ==================================== Graph Data Structure ====================================
# In tf_geometric, the data of a graph can be represented by either a collections of
# tensors (numpy.ndarray or tf.Tensor) or a tfg.Graph object.
# A graph usually consists of x(node features), edge_index and edge_weight(optional)

# Node Features => (num_nodes, num_features)
x = np.random.randn(5, 20).astype(np.float32)  # 5 nodes, 20 features

# Edge Index => (2, num_edges)
# Each column of edge_index (u, v) represents an directed edge from u to v.
# Note that it does not cover the edge from v to u. You should provide (v, u) to cover it.
# This is not convenient for users.
# Thus, we allow users to provide edge_index in undirected form and convert it later.
# That is, we can only provide (u, v) and convert it to (u, v) and (v, u) with `convert_edge_to_directed` method.
edge_index = np.array([
    [0, 0, 1, 3],
    [1, 2, 2, 1]
])

# Edge Weight => (num_edges)
edge_weight = np.array([0.9, 0.8, 0.1, 0.2]).astype(np.float32)

# Make the edge_index directed such that we can use it as the input of GCN
edge_index, [edge_weight] = convert_edge_to_directed(edge_index, [edge_weight])

print(tfg.nn.gcn_norm_edge(edge_index, 10))


# coding=utf-8
import tensorflow as tf
import numpy as np
import tf_sparse as tfs

def convert_union_to_numpy(data, dtype=None):
    if data is None:
        return data

    if tf.is_tensor(data):
        np_data = data.numpy()
    elif isinstance(data, list):
        np_data = np.array(data)
    else:
        np_data = data

    if dtype is not None:
        np_data = np_data.astype(dtype)

    return np_data


def union_len(data):
    if tf.is_tensor(data):
        return tfs.shape(data)[0]
    else:
        return len(data)


# coding=utf-8

# nn package contains functional APIs for tf_geometric


from .kernel.map_reduce import identity_mapper, neighbor_count_mapper, sum_reducer, sum_updater, identity_updater, mean_reducer, max_reducer, aggregate_neighbors
from .conv.gcn import gcn, gcn_norm_adj, gcn_build_cache_for_graph, gcn_norm_edge, gcn_cache_normed_edge
from .conv.gat import gat
from .conv.chebynet import chebynet, chebynet_norm_edge
from .conv.sgc import sgc
from .conv.tagcn import tagcn
from .conv.graph_sage import mean_graph_sage, sum_graph_sage, mean_pool_graph_sage, max_pool_graph_sage, gcn_graph_sage, lstm_graph_sage
from .conv.appnp import appnp
from .conv.gin import gin
from .conv.le_conv import le_conv
from .conv.ssgc import ssgc


from .sampling.drop_edge import drop_edge

from .pool.common_pool import mean_pool, min_pool, max_pool, sum_pool
from .pool.topk_pool import topk_pool
from .pool.diff_pool import diff_pool, diff_pool_coarsen
from .pool.set2set import set2set
from .pool.cluster_pool import cluster_pool
from .pool.sag_pool import sag_pool
from .pool.asap import asap
from .pool.sort_pool import sort_pool
from .pool.min_cut_pool import min_cut_pool, min_cut_pool_coarsen, min_cut_pool_compute_losses



# coding=utf-8
from .conv.gcn import GCN
from .conv.gat import GAT
from .conv.gin import GIN
from .conv.graph_sage import MeanGraphSage, SumGraphSage, MeanPoolGraphSage, MaxPoolGraphSage, GCNGraphSage, LSTMGraphSage
from .conv.sgc import SGC
from .conv.tagcn import TAGCN
from .conv.chebynet import ChebyNet
from .conv.appnp import APPNP
from .conv.le_conv import LEConv
from .conv.ssgc import SSGC


from .sampling.drop_edge import DropEdge

from .kernel.map_reduce import MapReduceGNN

from .pool.common_pool import MeanPool, MinPool, MaxPool, SumPool
from .pool.diff_pool import DiffPool
from .pool.set2set import Set2Set
from .pool.sag_pool import SAGPool
from .pool.asap import ASAP
from .pool.sort_pool import SortPool
from .pool.min_cut_pool import MinCutPool



# coding=utf-8
import os

from tf_geometric import SparseAdj
from tf_geometric.nn import gcn_norm_adj
from tf_sparse import SparseMatrix

os.environ["CUDA_VISIBLE_DEVICES"] = "5"
from tf_geometric.utils import tf_utils
import tensorflow as tf
import tf_geometric as tfg
from tqdm import tqdm
import time

# tf.config.run_functions_eagerly(True)

dataset = "ogbn-arxiv"
graph, (train_index, valid_index, test_index) = tfg.datasets.OGBNodePropPredDataset(dataset).load_data()
# graph: tfg.Graph

num_classes = graph.y.max() + 1
# graph.convert_data_to_tensor()

drop_rate = 0.5
learning_rate = 1e-2


# Multi-layer GCN Model
class GCNModel(tf.keras.Model):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.gcn0 = tfg.layers.GCN(256, activation=tf.nn.relu)
        self.gcn1 = tfg.layers.GCN(256, activation=tf.nn.relu)
        self.gcn2 = tfg.layers.GCN(num_classes)
        self.dropout = tf.keras.layers.Dropout(drop_rate)

    def call(self, inputs, training=None, mask=None, cache=None):
        # start = tf.timestamp()

        x, edge_index, edge_weight = inputs
        # h = self.dropout(x, training=training)
        h = self.gcn0([x, edge_index, edge_weight], cache=cache)
        h = self.dropout(h, training=training)

        h = self.gcn1([h, edge_index, edge_weight], cache=cache)
        h = self.dropout(h, training=training)
        h = self.gcn2([h, edge_index, edge_weight], cache=cache)

        # tf.print("forward time: ", tf.timestamp() - start)

        return h

# Multi-layer GCN Model
# class GCNModel1(tf.keras.Model):
#
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.gcn0 = tfg.layers.GCN(256, activation=tf.nn.relu)
#         self.gcn1 = tf.keras.layers.Dense(256, activation=tf.nn.relu)
#         self.gcn2 = tf.keras.layers.Dense(num_classes)
#         self.dropout = tf.keras.layers.Dropout(drop_rate)
#
#     def call(self, inputs, training=None, mask=None, cache=None):
#         start = tf.timestamp()
#
#         x, edge_index, edge_weight = inputs
#
#         h = self.gcn1(x, training=training)
#
#         adj: SparseAdj = cache["gcn_normed_adj_True_False"]
#         h = adj @ h
#
#         h = self.gcn2(h, training=training)
#
#
#
#         tf.print("forward time: ", tf.timestamp() - start)
#
#         return h

model = GCNModel()


# @tf_utils.function can speed up functions for TensorFlow 2.x.
# @tf_utils.function is not compatible with TensorFlow 1.x and dynamic graph.cache.
# @tf_utils.function
def forward(graph, training=False):
    return model([graph.x, graph.edge_index, graph.edge_weight], training=training, cache=graph.cache)




# The following line is only necessary for using GCN with @tf_utils.function
# For usage without @tf_utils.function, you can commont the following line and GCN layers can automatically manager the cache
model.gcn0.build_cache_for_graph(graph)

# adj: SparseAdj = graph.cache['gcn_normed_adj_True_False']

# print(tf.constant(adj))
# asdfasdf

# adj.index = adj.edge_index.numpy()
# adj.value = adj.edge_weight.numpy()

# adj.index = tf.constant(adj.edge_index)
# adj.value = tf.constant(adj.edge_weight)



# @tf_utils.function
def compute_loss(logits, mask_index, vars):
    # start = tf.timestamp()

    # masked_logits = tf.gather(logits, mask_index)
    # masked_labels = tf.gather(graph.y, mask_index)
    # # losses = tf.nn.softmax_cross_entropy_with_logits(
    # #     logits=masked_logits,
    # #     labels=tf.one_hot(masked_labels, depth=num_classes)
    # # )
    #
    # losses = tf.nn.sparse_softmax_cross_entropy_with_logits(
    #     logits=masked_logits,
    #     labels=masked_labels
    # )

    losses = tf.reduce_mean(logits)


    kernel_vars = [var for var in vars if "kernel" in var.name]
    l2_losses = [tf.nn.l2_loss(kernel_var) for kernel_var in kernel_vars]

    loss = tf.reduce_mean(losses) + tf.add_n(l2_losses) * 1e-4

    # tf.print("loss time: ", tf.timestamp() - start)

    return loss


optimizer = tf.keras.optimizers.Adam(learning_rate=learning_rate)


@tf_utils.function
# @tf_utils.function(experimental_compile=True)
def train_step():

    print("retracing ......")

    logits = forward(graph, training=True)
    loss = logits[0][0]#compute_loss(logits, train_index, model.trainable_variables)



    # with tf.init_scope():
    #     print("compute cache")
    #     model.gcn0.build_cache_for_graph(graph)

    # start = tf.timestamp()

    # with tf.GradientTape() as tape:
    #     logits = forward(graph, training=True)
    #     loss = compute_loss(logits, train_index, tape.watched_variables())

    # vars = tape.watched_variables()
    # grads = tape.gradient(loss, vars)
    # optimizer.apply_gradients(zip(grads, vars))

    # tf.print("train time: ", tf.timestamp() - start)

    return loss


# func = train_step.get_concrete_function()
# graph = func.graph
# for node in graph.as_graph_def().node:
#   print(f'{node.input} -> {node.name}')
#
#
# print("========")

@tf_utils.function
# @tf_utils.function(experimental_compile=True)
def evaluate():
    logits = forward(graph)
    masked_logits = tf.gather(logits, test_index)
    masked_labels = tf.gather(graph.y, test_index)

    y_pred = tf.argmax(masked_logits, axis=-1, output_type=tf.int32)

    corrects = tf.equal(y_pred, masked_labels)
    accuracy = tf.reduce_mean(tf.cast(corrects, tf.float32))
    return accuracy


loss = train_step()

# train_step_ = train_step.get_concrete_function()
# graph = train_step_.graph
# # for node in graph.as_graph_def().node:
# #   print(f'{node.input} -> {node.name}')
# #
# # print("======")
# # print(graph.as_graph_def())
#
# with open("numpy.txt", "w", encoding="utf-8") as f:
#     f.write(str(graph.as_graph_def()))
# exit(0)

accuracy = evaluate()

start_time = time.time()

print(train_step.pretty_printed_concrete_signatures())

for step in range(1, 20001):
    loss = train_step()
    if step % 20 == 0:
        # accuracy = evaluate()
        accuracy = 0.0
        print("step = {}\tloss = {}\taccuracy = {}\tavg_time = {}".format(step, loss, accuracy, (time.time() - start_time) / step))




print("\nstart speed test...")
num_test_iterations = 1000
start_time = time.time()
for _ in tqdm(range(num_test_iterations)):
    logits = forward(graph)
end_time = time.time()
print("mean forward time: {} seconds".format((end_time - start_time) / num_test_iterations))

if tf.__version__[0] == "1":
    print("** @tf_utils.function is disabled in TensorFlow 1.x. "
          "Upgrade to TensorFlow 2.x for 10X faster speed. **")


# coding=utf-8
import tensorflow as tf
from tf_geometric.nn.kernel.segment import segment_softmax

from tf_geometric.utils.union_utils import union_len


def set2set(x, node_graph_index, lstm, num_iterations, training=None):
    """
    Functional API for Set2Set

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param node_graph_index: Tensor/NDArray, shape: [num_nodes], graph index for each node
    :param lstm: A lstm model.
    :param num_iterations: Number of iterations for attention.
    :param training: Python boolean indicating whether the layer should behave in
        training mode (adding dropout) or in inference mode (doing nothing).
    :return: Graph features, shape: [num_graphs, num_node_features * 2]
    """

    num_graphs = tf.reduce_max(node_graph_index) + 1

    lstm_units = tf.shape(x)[-1]

    h = tf.zeros([num_graphs, lstm_units * 2], dtype=tf.float32)
    initial_state = [tf.zeros([1, lstm_units], dtype=tf.float32), tf.zeros([1, lstm_units], dtype=tf.float32)]

    for _ in range(num_iterations):

        h = tf.expand_dims(h, axis=0)
        h, state_h, state_c = lstm(h, initial_state=initial_state, training=training)
        initial_state = [state_h, state_c]
        h = tf.squeeze(h, axis=0)

        repeated_h = tf.gather(h, node_graph_index)
        # attention
        att_score = tf.reduce_sum(x * repeated_h, axis=-1, keepdims=True)
        normed_att_score = segment_softmax(att_score, node_graph_index, num_graphs)
        att_h = tf.math.unsorted_segment_sum(x * normed_att_score, node_graph_index, num_graphs)
        h = tf.concat([h, att_h], axis=-1)

    return h
